local MT = MacroToolkit
local _G = _G
MT.Spells = {}
local MTF, MTPF
local L = MT.L
local EditMacro, CreateMacro, GetMacroInfo, PickupMacro, DeleteMacro = EditMacro, CreateMacro, GetMacroInfo, PickupMacro, DeleteMacro
local GetSpellTexture, GetMacroIcons, GetMacroItemIcons = GetSpellTexture, GetMacroIcons, GetMacroItemIcons
local GetNumSpellTabs, GetSpellTabInfo, GetSpellBookItemInfo = GetNumSpellTabs, GetSpellTabInfo, GetSpellBookItemInfo
local GetSpellInfo, UnitName, UnitAura = GetSpellInfo, UnitName, UnitAura
local GetSpellBookItemTexture, GetFlyoutInfo, GetFlyoutSlotInfo = GetSpellBookItemTexture, GetFlyoutInfo, GetFlyoutSlotInfo
local FauxScrollFrame_GetOffset, FauxScrollFrame_Update = FauxScrollFrame_GetOffset, FauxScrollFrame_Update
local GetNumMacros, strsplit, date, table, floor = GetNumMacros, strsplit, date, table, floor
local string, ceil, max, mod, ipairs, select = string, ceil, max, mod, ipairs, select
local CreateFrame, GetContainerItemInfo, GetItemInfo, UseContainerItem = CreateFrame, GetContainerItemInfo, GetItemInfo, UseContainerItem
local PickupContainerItem, DeleteCursorItem, format, print = PickupContainerItem, DeleteCursorItem, format, print
MT.MACRO_ICON_FILENAMES = {}

MT.defaults = {
	profile = {
		override = false,
		stringcolour = "ffffffff",
		emotecolour = "ffeedd82",
		scriptcolour = "ff696969",
		commandcolour = "ff00bfff",
		spellcolour = "ff9932cc",
		targetcolour = "ffffd700",
		conditioncolour = "ff8b5a2b",
		defaultcolour = "ffffffff",
		errorcolour = "ffff0000",
		itemcolour = "fff08080",
		mtcolour = "ffcd2ea9",
		usecolours = true,
		showerrors = true,
		unknown = false,
		replacemt = true,
		x = (UIParent:GetWidth() - 638) / 2,
		y = (UIParent:GetHeight() - 424) / 2,
	},
	global = {
		custom = {},
		extended = {},
	},
	char = { extended = {} },
}

local function showtoolkit()
	local ignore = MacroToolkitFrame:IsShown() or MacroToolkitScriptFrame:IsShown() or MacroToolkitBuilderFrame:IsShown()
	if not ignore then ShowUIPanel(MacroToolkitFrame) end
end

function MT:eventHandler(this, event, arg1, ...)
	if event == "ADDON_LOADED" then
		if arg1 == "MacroToolkit" then
			_G.SLASH_MACROTOOLKIT_CMD1 = MT.slash .. "mac"
			_G.SLASH_MACROTOOLKIT_CMD2 = MT.slash .. "mtoolkit"
			_G.SLASH_MACROTOOLKIT_CMD3 = MT.slash .. "macrotoolkit"
			SlashCmdList["MACROTOOLKIT_CMD"] = showtoolkit
			for i, s in ipairs(MT.scripts) do
				_G["SLASH_MACROTOOLKIT_" .. string.upper(s[2]) .. 1] = MT.slash .. s[2]
				SlashCmdList["MACROTOOLKIT_" .. string.upper(s[2])] = function(input) MT:DoMTMacroCommand(s[2], input) end
			end
			MT.db = MT.LS("AceDB-3.0"):New("MacroToolkitDB", MT.defaults, "profile")
			if not IsAddOnLoaded("Blizzard_MacroUI") then LoadAddOn("Blizzard_MacroUI") end
			for _, c in ipairs(MT.db.global.custom) do
				_G["SLASH_MACROTOOLKIT_CUSTOM_" .. string.upper(c.n) .. "1"] = MT.slash .. c.n
				SlashCmdList["MACROTOOLKIT_CUSTOM_" .. string.upper(c.n)] = function(input) MT:DoCustomCommand(c.s, input) end
			end
		elseif arg1 == "Blizzard_MacroUI" then
			if MT.db.profile.override then
				MT.showmacroframe = ShowMacroFrame
				ShowMacroFrame = showtoolkit
			end
			local mtbutton = CreateFrame("Button", "MacroToolkitOpen", MacroFrame, "UIPanelButtonTemplate")
			mtbutton:SetText(L["Toolkit"])
			mtbutton:SetSize(94, 22)
			mtbutton:SetPoint("LEFT", MacroDeleteButton, "RIGHT")
			mtbutton:SetScript("OnClick",
				function()
					ShowUIPanel(MacroToolkitFrame)
					HideUIPanel(MacroFrame)
				end)
		end
	elseif event == "PLAYER_LOGIN" then
		MT:BuildCommandList()
		MT:CreateOptions()
		MTF = MT:CreateMTFrame()
		MT:SetMacros(true)
		MTPF = MT:CreateMTPopup()
		MT:CreateRestoreFrame()
		MT:CreateScriptFrame()
		MT:CreateSecureFrames()
		MT:CreateBuilderFrame()
		for i, e in pairs(MT.db.global.extended) do _G["MacroToolkitSecureButton" .. i]:SetAttribute("macrotext", e) end
		for i, e in pairs(MT.db.char.extended) do _G["MacroToolkitSecureButton" .. i]:SetAttribute("macrotext", e) end
		if not MT.db.profile.usecolours then MacroToolkitFauxScrollFrame:Hide() end
		if not MT.db.profile.showerrors then
			MacroToolkitErrorBg:Hide()
			MacroToolkitErrorScrollFrame:Hide()
		end
		UIPanelWindows.MacroToolkitFrame = { area = "center", pushable = 0, whileDead = 1, enabled = 1}
		UIPanelWindows.MacroToolkitScriptFrame = { area = "center", pushable = 0, whileDead = 1, enabled = 1}
		HideUIPanel(MTF)
		if IsAddOnLoaded("ElvUI") then MT:LoadElvSkin() end
	end
end

function MT:GetBackups()
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
	local var = (tab == 1) and "global" or "char"
	local blist = {}
	for _, d in ipairs(MT.db[var].backups) do blist[d.d] = d.d end
	return blist
end

function MT:GetBackupTitle(key)
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
	local var = (tab == 1) and "global" or "char"
	for _, d in ipairs(MT.db[var].backups) do if d.d == key then return d.n end end
end

function MT:ClearAllMacros()
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
	local mstart = (tab == 1) and 1 or (_G.MAX_ACCOUNT_MACROS + 1)
	local mend = (tab == 1) and _G.MAX_ACCOUNT_MACROS or (_G.MAX_ACCOUNT_MACROS + _G.MAX_CHARACTER_MACROS)
	for m = mend, mstart, -1 do
		DeleteMacro(m)
		MT:MacroFrameUpdate()
		_G["MacroToolkitSecureButton" .. m]:SetAttribute("macrotext", "")
	end
	local var = (tab == 1) and "global" or "char"
	MT.db[var].extended = {}
	MT:HideDetails()
	MacroToolkitText:SetText("")
	MacroToolkitFauxText:SetText("")
	MTF.selectedMacro = nil
	MacroToolkitSelMacroName:SetText("")
end

function MT:DeleteBackup(dd, desc)
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
	local var = (tab == 1) and "global" or "char"
	for i, d in ipairs(MT.db[var].backups) do
		if d.d == MT.backup then
			table.remove(MT.db[var].backups, i)
			MT:GetLastBackupDate()
			dd:SetValue("")
			desc:SetText("")
			local blist = MT:GetBackups()
			dd:SetList(blist)
			if #blist == 0 then
				MacroToolkitRestoreRestore:Disable()
				MacroToolkitRestoreDelete:Disable()
			end
			break
		end
	end
end

function MT:HexToRGB(hex)
    local ahex, rhex, ghex, bhex = string.sub(hex, 1, 2), string.sub(hex, 3, 4), string.sub(hex, 5, 6), string.sub(hex, 7, 8)
	return tonumber(rhex, 16) / 255, tonumber(ghex, 16) / 255, tonumber(bhex, 16) / 255, tonumber(ahex, 16) / 255
end

function MT:RGBToHex(r, g, b, a)
	r = r <= 1 and r >= 0 and r or 0
	g = g <= 1 and g >= 0 and g or 0
	b = b <= 1 and b >= 0 and b or 0
	a = a <= 1 and a >= 0 and a or 0
	return string.format("%.2x%.2x%.2x%.2x", a*255, r*255, g*255, b*255)
end

function MT:DoCustomCommand(luatext, input)
	local a1, a2, a3, a4 = strsplit(" ", input)
	if not a1 then a1, a2, a3, a4 = strsplit(",", input) end
	local script = string.format("local arg1, arg2, arg3, arg4 = \"%s\", \"%s\", \"%s\", \"%s\"\n%s", a1 or "", a2 or "", a3 or "", a4 or "", luatext)
	assert(loadstring(script))()
end

function MT:DoMTMacroCommand(command, parameters)
	if command == "mtce" then UIErrorsFrame:Clear()
	elseif command == "mtrp" then SummonRandomCritter()
	elseif command == "mtso" then SetCVar("Sound_EnableSFX", 1)
	elseif command == "mtsx" then SetCVar("Sound_EnableSFX", 0)
	elseif command == "mtrt" then
		local t, m = strsplit(" ", parameters)
		SetRaidTarget(t, m)
	elseif command == "mtev" then VehicleExit()
	elseif command == "mtmc" then
		local x, y = GetPlayerMapPosition("player")
		print(format("%.1f %.1f", x*100, y*100))
	elseif command == "mttc" then ShowCloak(not ShowingCloak())
	elseif command == "mtth" then ShowHelm(not ShowingHelm())
	elseif command == "mtep" then for seat = 1,2 do if CanEjectPassengerFromSeat(seat) then EjectPassengerFromSeat(seat) end end
	elseif command == "mtsg" then
		local GetContainerItemInfo, GetItemInfo, UseContainerItem, cash = GetContainerItemInfo, GetItemInfo, UseContainerItem, 0
		for bag = 0, 4 do
			for slot = 1, GetContainerNumSlots(bag) do
				local _, qty, _, _, _, _, itemlink = GetContainerItemInfo(bag, slot)
				if string.find(itemlink or "", "9d9d9d") then
					local vendorprice = select(11, GetItemInfo(itemlink)) * qty
					cash = cash + vendorprice
					UseContainerItem(bag, slot)
				end
			end
		end
		local gold = floor(cash / (_G.COPPER_PER_SILVER * _G.SILVER_PER_GOLD))
		local goldDisplay = BreakUpLargeNumbers(gold)
		local silver = floor((cash - (gold * _G.COPPER_PER_SILVER * _G.SILVER_PER_GOLD)) / _G.COPPER_PER_SILVER)
		local copper = mod(cash, _G.COPPER_PER_SILVER)
		local goldtexture = "|TInterface\\MoneyFrame\\UI-GoldIcon:12:12|t"
		local silvertexture = "|TInterface\\MoneyFrame\\UI-SilverIcon:12:12|t"
		local coppertexture = "|TInterface\\MoneyFrame\\UI-CopperIcon:12:12|t"
		print(L["Sell grey items"] .. ": " .. gold .. goldtexture .. silver .. silvertexture .. copper .. coppertexture)
	elseif command == "mtdg" then
		local GetContainerItemInfo, GetItemInfo, PickupContainerItem, DeleteCursorItem = GetContainerItemInfo, GetItemInfo, PickupContainerItem, DeleteCursorItem
		for bag = 0, 4 do
			for slot = 1, GetContainerNumSlots(bag) do
				local itemlink = select(7, GetContainerItemInfo(bag, slot))
				if string.find(itemlink or "", "9d9d9d") then
					PickupContainerItem(bag, slot)
					DeleteCursorItem()
				end
			end
		end
	elseif command == "mtnb" then
		local UnitAura, UnitName = UnitAura, UnitName
		local msg = L["No food buff"] .. ": "
		local ppl = _G.NONE
		for member = 1, GetNumGroupMembers() do
			for buff = 1, 41 do
				local aura = UnitAura("raid" .. member, buff)
				if aura == L["Well Fed"] or aura == L["Food"] then break
				elseif buff == 41 and aura ~= L["Well Fed"] then ppl = ppl .. UnitName("raid" .. member) .. " " end
			end
		end
		SendChatMessage(msg .. ppl, MT.channel or "raid")
	elseif command == "mtnf" then
		local UnitAura, UnitName = UnitAura, UnitName
		local msg = L["No food buff"] .. ": "
		local ppl = _G.NONE
		for member = 1, GetNumGroupMembers() do
			for buff = 1, 41 do
				local aura = UnitAura("raid" .. member, buff)
				if aura then
					if string.find(aura, L["Flask"])or string.find(aura, L["Distilled"])then break end
				elseif buff == 41 then ppl = ppl .. UnitName("raid" .. member) .. " " end
			end
		end
		SendChatMessage(msg .. ppl, "raid")
	end
end

function MT:FindScript(scriptname)
	for _, s in ipairs(MT.scripts) do
		if s[1] == scriptname then return s end
	end
end

function MT:ShowShortened(chars)
	StaticPopupDialogs["MACROTOOLKIT_SHORTENED"] = {
		text = string.format((chars == 1) and L["Macro shortened by %d character"] or L["Macro shortened by %d characters"], chars),
		button1 = _G.OKAY,
		timeout = 0,
		exclusive = 1,
		whileDead = 1,
		hideOnEscape = 1
	}
	StaticPopup_Show("MACROTOOLKIT_SHORTENED")
end

function MT:SetMacros(account)
	MTF.macroBase = account and 0 or _G.MAX_ACCOUNT_MACROS
	MTF.macroMax = account and _G.MAX_ACCOUNT_MACROS or _G.MAX_CHARACTER_MACROS
	local numAccountMacros, numCharacterMacros = GetNumMacros()
	if (account and numAccountMacros or numCharacterMacros) > 0 then MT:SelectMacro(MTF.macroBase + 1)
	else MT:SelectMacro(nil) end
end

function MT:FormatMacro(macrotext)
	if macrotext == "" then return "", "" end
	local lines = {strsplit("\n", macrotext)}
	local out, eout = "", ""
	for n, l in ipairs(lines) do
		if l == "" then out = out .. "\n"
		else
			local f, err = MT:ParseMacro(l)
			if not f then break end
			out = out .. f .. "\n"
			for _, e in ipairs(err) do eout = eout .. e .. "\n" end
		end
	end
	out = string.sub(out, 1, string.len(out) - 1)
	return out, eout
end

local function updateextended(index, body, mindex)
	local var = (index > _G.MAX_ACCOUNT_MACROS) and "char" or "global"
	local mindex = tostring(mindex)
	if not MT.db[var].extended then MT.db[var].extended = {} end
	MT.db[var].extended[mindex] = body
end

function MT:DeleteExtended(index)
	local var = (MTF.selectedMacro > _G.MAX_ACCOUNT_MACROS) and "char" or "global"
	index = tostring(index)
	MT.db[var].extended[index] = nil
end

function MT:ExtendClick(this)
	local extend = this:GetText() == L["Extend"]
	if extend then MT:ExtendMacro()
	else MT:UnextendMacro() end
end

function MT:GetExtendedNextIndex()
	for m = 1, _G.MAX_ACCOUNT_MACROS + _G.MAX_CHARACTER_MACROS do
		local var = (m > _G.MAX_ACCOUNT_MACROS) and "char" or "global"
		if not MT.db[var].extended[tostring(m)] then return tostring(m) end
	end
	return "1"
end

function MT:GetCurrentIndex()
	if not MTF.selectedMacro then return end
	local _, _, body = GetMacroInfo(MTF.selectedMacro)
	local _, _, index = string.find(body, "MacroToolkitSecureButton(%d+)")
	return index
end

function MT:ExtendMacro(save)
	local body = MacroToolkitText:GetText()
	local index = save and MT:GetCurrentIndex() or MT:GetExtendedNextIndex()
	--if not index then return end
	local securebutton = _G["MacroToolkitSecureButton" .. index]
	updateextended(MTF.selectedMacro, body, index)
	securebutton:SetAttribute("macrotext", body)
	if not save then
		local newbody = MT.click .. " " .. securebutton:GetName()
		MacroToolkitText.extended = true
		_G["MacroToolkitButton" .. MTF.selectedMacro].extended = true
		EditMacro(MTF.selectedMacro, nil, nil, newbody)
		MacroToolkitExtend:SetText(L["Unextend"])
		MacroToolkitText:GetScript("OnTextChanged")(MacroToolkitText)
		MT:UpdateCharLimit()
	end
end

local function unextend(body)
	local index = MTF.selectedMacro
	local _, _, body = GetMacroInfo(index)
	local _, _, mindex = string.find(body, "MacroToolkitSecureButton(%d+)")
	local securebutton = _G["MacroToolkitSecureButton" .. mindex]
	local macrotext = securebutton:GetAttribute("macrotext")
	MT:DeleteExtended(mindex)
	securebutton:SetAttribute("macrotext", "")
	MacroToolkitText.extended = nil
	_G["MacroToolkitButton" .. MTF.selectedMacro].extended = nil
	MacroToolkitExtend:SetText(L["Extend"])
	EditMacro(MTF.selectedMacro, nil, nil, macrotext)
	MT:SaveMacro()
	MT:MacroFrameUpdate()
	MT:UpdateCharLimit()
end

function MT:UnextendMacro()
	local body = MacroToolkitText:GetText()
	if string.len(body) > 255 then
		StaticPopupDialogs.MACROTOOLKIT_DELETEBACKUP.text = L["Your macro will be truncated to 255 characters. Are you sure?"]
		StaticPopupDialogs.MACROTOOLKIT_DELETEBACKUP.OnAccept = function() unextend(string.sub(body, 1, 255)) end
		StaticPopup_Show("MACROTOOLKIT_DELETEBACKUP")
	else unextend(body) end
end

function MT:GetExtendedBody(index, tab)
	local var = (tab == 1) and "global" or "char"
	return MT.db[var].extended[index]
end

function MT:UpdateCharLimit()
	local extended = MacroToolkitText.extended
	local dct = MacroToolkitText:GetText()
	local chars = string.len(dct)
	local ltt = extended and L["%d characters used"] or _G.MACROFRAME_CHAR_LIMIT
	if chars > 255 and not extended then
		local ft = string.sub(dct, 1, 255)
		MacroToolkitText:SetText(ft)
		MacroToolkitLimit:SetFormattedText(ltt, 255)
	else MacroToolkitLimit:SetFormattedText(ltt, chars) end
end

function MT:MacroFrameUpdate()
	local numMacros
	local numAccountMacros, numCharacterMacros = GetNumMacros()
	local macroButtonName, macroButton, macroIcon, macroName
	local name, texture, body
	local selectedName, selectedBody, selectedIcon
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)

	numMacros = (MTF.macroBase == 0) and numAccountMacros or numCharacterMacros

	-- Macro List
	local maxMacroButtons = max(_G.MAX_ACCOUNT_MACROS, _G.MAX_CHARACTER_MACROS)
	for i = 1, maxMacroButtons do
		macroButtonName = "MacroToolkitButton" .. i
		macroButton = _G[macroButtonName]
		macroIcon = _G[macroButtonName .. "Icon"]
		macroName = _G[macroButtonName .. "Name"]
		if i <= MTF.macroMax then
			if i <= numMacros then
				name, texture, body = GetMacroInfo(MTF.macroBase + i)
				macroIcon:SetTexture(texture)
				macroName:SetText(name)
				macroButton:Enable()
				-- Highlight Selected Macro
				if MTF.selectedMacro and i == (MTF.selectedMacro - MTF.macroBase) then
					macroButton:SetChecked(1)
					MacroToolkitSelMacroName:SetText(name)
					local s, e, index = string.find(body, "MacroToolkitSecureButton(%d+)")
					if index then
						body = MT:GetExtendedBody(index, tab)
						MacroToolkitText.extended = true
						MacroToolkitExtend:SetText(L["Unextend"])
						macroButton.extended = true
					else
						MacroToolkitText.extended = nil
						macroButton.extended = nil
						MacroToolkitExtend:SetText(L["Extend"])
					end
					body = body or ""
					local m, e = MT:FormatMacro(body)
					MacroToolkitText:SetText(body)
					MacroToolkitFauxText:SetText(m)
					MacroToolkitErrors:SetText(e)
					MacroToolkitSelMacroButton:SetID(i)
					MacroToolkitSelMacroButtonIcon:SetTexture(texture)
					if MTPF then MTPF.selectedIconTexture = string.gsub(string.upper(texture), "INTERFACE\\ICONS\\", "") end
				else macroButton:SetChecked(0) end
			else
				macroButton:SetChecked(0)
				macroIcon:SetTexture("")
				macroName:SetText("")
			end
			macroButton:Show()
		else macroButton:Hide() end
	end

	-- Macro Details
	if MTF.selectedMacro ~= nil then
		MT:ShowDetails()
		MacroToolkitDelete:Enable()
		MacroToolkitShorten:Enable()
		MacroToolkitExtend:Enable()
		MacroToolkitBackup:Enable()
		MacroToolkitShare:Enable()
		MacroToolkitConditions:Enable()
	else
		MT:HideDetails()
		MacroToolkitDelete:Disable()
		MacroToolkitShorten:Disable()
		MacroToolkitExtend:Disable()
		MacroToolkitBackup:Disable()
		MacroToolkitShare:Disable()
		MacroToolkitConditions:Disable()
	end
	
	--Update New Button
	if numMacros < MTF.macroMax then MacroToolkitNew:Enable()
	else MacroToolkitNew:Disable() end

	-- Disable Buttons
	if MTPF then
		if MTPF:IsShown() then
			MacroToolkitEdit:Disable()
			MacroToolkitDelete:Disable()
			MacroToolkitShorten:Disable()
			MacroToolkitExtend:Disable()
			MacroToolkitBackup:Disable()
			MacroToolkitClear:Disable()
			MacroToolkitShare:Disable()
			MacroToolkitConditions:Disable()
		else
			MacroToolkitEdit:Enable()
			MacroToolkitDelete:Enable()
			MacroToolkitShorten:Enable()
			MacroToolkitExtend:Enable()
			MacroToolkitBackup:Enable()
			MacroToolkitClear:Enable()
			MacroToolkitShare:Enable()
			MacroToolkitConditions:Enable()
		end
	end

	if not MTF.selectedMacro then
		MacroToolkitDelete:Disable()
		MacroToolkitShorten:Disable()
		MacroToolkitExtend:Disable()
		MacroToolkitShare:Disable()
	end
	
	if numMacros > 0 then MacroToolkitClear:Enable()
	else MacroToolkitClear:Disable() end
end

function MT:ContainerOnLoad(this)
	local button
	local maxMacroButtons = max(_G.MAX_ACCOUNT_MACROS, _G.MAX_CHARACTER_MACROS)
	for i = 1, maxMacroButtons do
		button = CreateFrame("CheckButton", "MacroToolkitButton" .. i, this, "MacroButtonTemplate")
		button:SetScript("OnClick", function(this, button) MT:MacroButtonOnClick(this, button) end)
		button:SetScript("OnDragStart", function(this) PickupMacro(MTF.macroBase + this:GetID()) end)
		button:SetID(i)
		if i == 1 then button:SetPoint("TOPLEFT", this, "TOPLEFT", 6, -6)
		elseif mod(i, _G.NUM_MACROS_PER_ROW) == 1 then button:SetPoint("TOP", "MacroToolkitButton" .. (i - _G.NUM_MACROS_PER_ROW), "BOTTOM", 0, -10)
		else button:SetPoint("LEFT", "MacroToolkitButton" .. (i - 1), "RIGHT", 13, 0) end
	end
end

function MT:MacroButtonOnClick(this, button)
	MT:SaveMacro()
	MT:SelectMacro(MTF.macroBase + this:GetID())
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
	local name = GetMacroInfo(this:GetID() + ((tab == 1) and 0 or _G.MAX_ACCOUNT_MACROS))
	if not name then
		if CursorHasMacro() then
			local _, mid = GetCursorInfo()
			local name, texture, body = GetMacroInfo(mid)
			local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
			texture = string.gsub(string.upper(texture), "INTERFACE\\ICONS\\", "")
			local s, e, index = string.find(body, "MacroToolkitSecureButton(%d+)")
			if s then
				local bg, bc = MT.db.global.extended[index], MT.db.char.extended[index]
				if bg and tab == 2 then
					MT.db.char.extended[index] = bg
					MT.db.global.extended[index] = nil
				elseif bc and tab == 1 then
					MT.db.global.extended[index] = bc
					MT.db.char.extended[index] = nil
				end
			end
			CreateMacro(name, texture, body, tab == 2)
			MT:MacroFrameUpdate()
			DeleteMacro(mid)
			ClearCursor()
		else
			MT:SelectMacro(nil)
			MT:HideDetails()
		end
	end
	MacroToolkitText.extended = button.extended
	MT:MacroFrameUpdate()
	MTPF:Hide()
	MacroToolkitText:ClearFocus()
end

function MT:SelectMacro(id) MTF.selectedMacro = id end

function MT:HideDetails()
	MacroToolkitEdit:Hide()
	MacroToolkitLimit:Hide()
	MacroToolkitText:Hide()
	MacroToolkitFauxText:Hide()
	MacroToolkitSelMacroName:Hide()
	MacroToolkitSelBg:Hide()
	MacroToolkitSelMacroButton:Hide()
end

function MT:ShowDetails()
	MacroToolkitEdit:Show()
	MacroToolkitLimit:Show()
	MacroToolkitEnterText:Show()
	MacroToolkitText:Show()
	MacroToolkitFauxText:Show()
	MacroToolkitSelMacroName:Show()
	MacroToolkitSelBg:Show()
	MacroToolkitSelMacroButton:Show()
	MacroToolkitShorten:Enable()
	MacroToolkitExtend:Enable()
end

function MT:RefreshPlayerSpellIconInfo()
	MT.MACRO_ICON_FILENAMES = {}
	MT.MACRO_ICON_FILENAMES[1] = "INV_MISC_QUESTIONMARK"
	local index = 2
	local numFlyouts = 0

	for i = 1, GetNumSpellTabs() do
		local tab, tabTex, offset, numSpells, _ = GetSpellTabInfo(i)
		offset = offset + 1
		local tabEnd = offset + numSpells
		for j = offset, tabEnd - 1 do
			local spellType, ID = GetSpellBookItemInfo(j, "player")
			if spellType ~= "FUTURESPELL" then
				local spellTexture = string.upper(GetSpellBookItemTexture(j, "player"))
				local spellName = GetSpellInfo(ID)
				if not string.match(spellTexture, "INTERFACE\\BUTTONS\\") then
					MT.MACRO_ICON_FILENAMES[index] = string.gsub(spellTexture, "INTERFACE\\ICONS\\", "")
					index = index + 1
				end
			end
			if spellType == "FLYOUT" then
				local _, _, numSlots, isKnown = GetFlyoutInfo(ID)
				if isKnown and (numSlots > 0) then
					for k = 1, numSlots do 
						local spellID, overrideSpellID, isKnown = GetFlyoutSlotInfo(ID, k)
						if isKnown then
							MT.MACRO_ICON_FILENAMES[index] = string.gsub(string.upper(GetSpellTexture(spellID)), "INTERFACE\\ICONS\\", "")
							index = index + 1
						end
					end
				end
			end
		end
	end
	GetMacroIcons(MT.MACRO_ICON_FILENAMES)
	GetMacroItemIcons(MT.MACRO_ICON_FILENAMES)
end

function MT:GetSpellorMacroIconInfo(index)
	if not index then return end
	return MT.MACRO_ICON_FILENAMES[index]
end

function MT:PopupUpdate(this)
	this = this or MTPF
	local numMacroIcons = #MT.MACRO_ICON_FILENAMES
	local macroPopupIcon, macroPopupButton
	local macroPopupOffset = FauxScrollFrame_GetOffset(MacroToolkitPopupScroll)
	local index
	
	if this.mode == "new" then MacroToolkitPopupEdit:SetText("")
	elseif this.mode == "edit" then
		local name, _, body = GetMacroInfo(MTF.selectedMacro)
		MacroToolkitPopupEdit:SetText(name)
	end

	local texture
	for i = 1, 30 do
		macroPopupIcon = _G["MacroToolkitPopupButton" .. i .. "Icon"]
		macroPopupButton = _G["MacroToolkitPopupButton" .. i]
		index = (macroPopupOffset * _G.NUM_ICONS_PER_ROW) + i
		texture = MT:GetSpellorMacroIconInfo(index)
		if index <= numMacroIcons and texture then
			macroPopupIcon:SetTexture("INTERFACE\\ICONS\\" .. texture)
			macroPopupButton:Show()
		else
			macroPopupIcon:SetTexture("")
			macroPopupButton:Hide()
		end
		if MTPF.selectedIcon and index == MTPF.selectedIcon then macroPopupButton:SetChecked(1)
		elseif MTPF.selectedIconTexture == texture then macroPopupButton:SetChecked(1)
		else macroPopupButton:SetChecked(nil) end
	end
	FauxScrollFrame_Update(MacroToolkitPopupScroll, ceil(numMacroIcons / _G.NUM_ICONS_PER_ROW) , 5, _G.MACRO_ICON_ROW_HEIGHT )
end

function MT:CancelEdit()
	MTPF:Hide()
	MT:MacroFrameUpdate()
	MTPF.selectedIcon = nil
end

function MT:PopupOkayUpdate()
	local text = MacroToolkitPopupEdit:GetText()
	text = string.gsub(text, "\"", "")
	if string.len(text) > 0 and MTPF.selectedIcon then MacroToolkitPopupOk:Enable()
	else MacroToolkitPopupOk:Disable() end
	if MTPF.mode == "edit" and string.len(text) > 0 then MacroToolkitPopupOk:Enable() end
end

function MT:SelectTexture(selectedIcon)
	MTPF.selectedIcon = selectedIcon
	MTPF.selectedIconTexture = nil
	MacroToolkitSelMacroButtonIcon:SetTexture("INTERFACE\\ICONS\\" .. MT:GetSpellorMacroIconInfo(MTPF.selectedIcon))
	MT:PopupOkayUpdate()
	local mode = MTPF.mode
	MTPF.mode = nil
	MT:PopupUpdate(MTPF)
	MTPF.mode = mode
end

function MT:PopupButtonOnClick(this) MT:SelectTexture(this:GetID() + (FauxScrollFrame_GetOffset(MacroToolkitPopupScroll) * _G.NUM_ICONS_PER_ROW)) end

function MT:PopupOkayButtonOnClick(this)
	local index = 1
	local iconTexture = MT:GetSpellorMacroIconInfo(MTPF.selectedIcon)
	local text = MacroToolkitPopupEdit:GetText()
	text = string.gsub(text, "\"", "")
	if MTPF.mode == "new" then index = CreateMacro(text, iconTexture, nil, (MTF.macroBase > 0))
	elseif MTPF.mode == "edit" then index = EditMacro(MTF.selectedMacro, text, iconTexture) end
	MTPF:Hide()
	MT:SelectMacro(index)
	MT:MacroFrameUpdate()
end

function MT:SaveMacro()
	if MTF.textChanged and MTF.selectedMacro then
		if MacroToolkitText.extended then MT:ExtendMacro(true)
		else EditMacro(MTF.selectedMacro, nil, nil, MacroToolkitText:GetText()) end
		MTF.textChanged = nil
	end
end

local function getMacroIcon(texture) for i, t in ipairs(MT.MACRO_ICON_FILENAMES) do if string.upper(t) == texture then return i end end end

function MT:GetBackupName()
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
	local mtype = _G["MacroToolkitFrameTab" .. tab]:GetText()
	StaticPopupDialogs.MACROTOOLKIT_BACKUPNAME.text = "|cffeedd82" .. mtype .. "|r\n\n" .. L["Enter a name for this backup"]
	StaticPopup_Show("MACROTOOLKIT_BACKUPNAME")
end

function MT:SetBackupName(dialog)
	local name = dialog.editBox:GetText() or ""
	dialog:Hide()
	MT:BackupMacros(name)
end

function MT:BackupMacros(backupname)
	local name, texture, body, fname
	local macros = {n=backupname, d=date(L["datetime format"]), m={}}
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
	local var = (tab == 1) and "global" or "char"
	local start = (tab == 1) and 1 or (_G.MAX_ACCOUNT_MACROS + 1)
	local finish = (tab == 1) and _G.MAX_ACCOUNT_MACROS or (_G.MAX_ACCOUNT_MACROS + _G.MAX_CHARACTER_MACROS)
	MT:RefreshPlayerSpellIconInfo()
	for m = start, finish do
		name, texture, body = GetMacroInfo(m)
		if name then
			if string.find(body, "MacroToolkitSecureButton") then body = _G["MacroToolkitSecureButton" .. m]:GetAttribute("macrotext") end
			fname = string.gsub(string.upper(texture), "INTERFACE\\ICONS\\", "")
			table.insert(macros.m, {index = m, icon = fname, body = body, name = name})
		end
	end
	if not MT.db[var].backups then MT.db[var].backups = {} end
	table.insert(MT.db[var].backups, macros)
	MT:SetLastBackupDate()
	MT:GetLastBackupDate()
	MT.MACRO_ICON_FILENAMES = nil
	collectgarbage()
end

local function createextended(name, icon, body, tab)
	local holder = "#MacroToolkitHolder"
	CreateMacro(name, icon, holder, tab == 2)
	local mid, mindex
	local var = (tab == 1) and "global" or "char"
	local start = (tab == 1) and 1 or (_G.MAX_ACCOUNT_MACROS + 1)
	local finish = (tab == 1) and _G.MAX_ACCOUNT_MACROS or (_G.MAX_ACCOUNT_MACROS + _G.MAX_CHARACTER_MACROS)
	for i, m in pairs(MT.db[var].extended) do
		if m == body then
			mindex = i
			break
		end
	end
	mindex = mindex or MT:GetExtendedNextIndex()
	for m = start, finish do
		local _, _, mbody = GetMacroInfo(m)
		if mbody == holder then
			mid = m
			break
		end
	end
	local securebutton = _G["MacroToolkitSecureButton" .. mindex]
	updateextended(start, body, mindex)
	securebutton:SetAttribute("macrotext", body)
	local newbody = MT.click .. " " .. securebutton:GetName()
	EditMacro(mid, nil, nil, newbody)
end

function MT:RestoreBackup()
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
	local var = (tab == 1) and "global" or "char"
	for i, d in ipairs(MT.db[var].backups) do
		if d.d == MT.backup then
			MT:RefreshPlayerSpellIconInfo()
			for _, m in ipairs(d.m) do
				if string.len(m.body) > 255 then createextended(m.name, m.icon, m.body, tab)
				else CreateMacro(m.name, m.icon, m.body, m.index > _G.MAX_ACCOUNT_MACROS) end
			end
			MT:MacroFrameUpdate()
			break
		end
	end
	MacroToolkitRestoreFrame:Hide()
end

function MT:GetLastBackupDate()
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
	local var = (tab == 1) and "global" or "char"
	if not MT.db[var] then MT.db[var] = {} end
	if not MT.db[var].backups then MT.db[var].backups = {} end
	local bd = MT.db[var].lastbackup or _G.NEVER
	if #MT.db[var].backups == 0 then bd = _G.NEVER end
	if bd == _G.NEVER then MacroToolkitRestore:Disable()
	else MacroToolkitRestore:Enable() end
	MacroToolkitBL:SetText(L["Last backup"] .. ": " .. bd)
end

function MT:SetLastBackupDate()
	local tab = PanelTemplates_GetSelectedTab(MacroToolkitFrame)
	local var = (tab == 1) and "global" or "char"
	MT.db[var].lastbackup = date(L["datetime format"])
end

function MT:BuildAndInsert()
	local b = MT.builder or {}
	local cargs = {}
	local failed
	local conditions =  b.target or ""
	if string.len(conditions) > 0 then conditions = "@" .. conditions .. "," end
	for to = 1, 4 do
		local opt = b["targetoption" .. to]
		if opt then conditions = conditions .. opt .. "," end
	end
	for o = 1, 5 do
		local opt = b["add" .. o]
		local arg = b["addp" .. o] or ""
		if opt then
			if (MT.conditions[opt] or 0 > 0) and (string.len(arg) > 0) then
				table.insert(cargs, {o = opt, a = arg})
			else conditions = conditions .. opt .. "," end
		end
	end
	table.sort(cargs, function(a, b) return a.o < b.o end)
	local lastarg
	for _, o in ipairs(cargs) do
		if o.o ~= lastarg then conditions = conditions .. o.o .. ":" .. o.a .. ","
		else conditions = conditions .. "/" .. o.a .. "," end
		lastarg = o.o
	end
	conditions = string.gsub(conditions, ",+", ",")
	conditions = string.gsub(conditions, ",/", "/")
	local clen = string.len(conditions)
	if string.sub(conditions, clen, clen) == "," then conditions = string.sub(conditions, 1, clen - 1) end
	conditions = "[" .. conditions .. "]"
	clen = string.len(conditions)
	mlen = string.len(MacroToolkitText:GetText())
	print("cond:",conditions,"clen:",clen,"mlen:",mlen)
	if not MacroToolkitText.extended then
		if clen + mlen > 255 then
			StaticPopupDialogs["MACROTOOLKIT_TOOLONG"].text = string.format(L["Not enough space. Command requires %d characters (%d available)"], clen, 255 - mlen)
			StaticPopup_Show("MACROTOOLKIT_TOOLONG")
			failed = true
		end
	end
	if not failed then
		MacroToolkitText:Insert(conditions)
		MacroToolkitText:GetScript("OnTextChanged")(MacroToolkitText)
		MacroToolkitFrame.textChanged = 1
		MT:SaveMacro()
		HideUIPanel(MacroToolkitBuilderFrame)
	end
end